"""
Data model for supported services
"""

services = [
    "accounts",
    "campaigns",
    "contacts",
    "creatives",
    "custom_fields",
    "external_users",
    "groups",
    "line_items",
    "line_item_creative_associations",
    "opportunities",
    "opportunity_line_items",
    "packages",
    "products",
    "product_rates",
    "rate_cards",
    "reports",
    "users",
]
