"""
Data model for the report service columns
"""

COLUMNS = [
    "ad_server",
    "ad_server_id",
    "first_party_impressions",
    "first_party_clicks",
    "first_party_ctr",
    "first_party_delivered_amount",
    "third_party_impressions",
    "third_party_clicks",
    "third_party_ctr",
    "third_party_delivered_amount",
    "reconciled_units",
    "reconciled_amount",
    "billed_units",
    "billed_amount",
]
